package com.pyx4me.cldcunit;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javassist.CannotCompileException;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewConstructor;
import javassist.CtNewMethod;
import javassist.NotFoundException;
import net.sf.jour.InterceptorException;
import net.sf.jour.filter.Pointcut;
import net.sf.jour.filter.PointcutListFilter;
import net.sf.jour.instrumentor.Instrumentor;
import net.sf.jour.instrumentor.InstrumentorResults;
import net.sf.jour.instrumentor.InstrumentorResultsImpl;
import net.sf.jour.log.Logger;

public class TestCaseReflectionDataInstrumentor implements Instrumentor {

	protected static final Logger log = Logger.getLogger(TestCaseReflectionDataInstrumentor.class);

	protected PointcutListFilter pointcuts;

	private static final String instrumentedFiled = "__testMethodsArray";

	public InstrumentorResults instrument(CtClass clazz) throws InterceptorException {

		if (clazz.isInterface() || clazz.getName().startsWith("junit.framework.")) {
			return InstrumentorResultsImpl.NOT_MODIFIED;
		}

		try {
			clazz.getField(instrumentedFiled);
			log.debug("already instrumented:" + clazz.getName());
			return InstrumentorResultsImpl.NOT_MODIFIED;
		} catch (NotFoundException ok) {
		}

		long countMethods = 0;

		boolean modified = false;

		List createdClasses = new Vector();
		HashMap instrumented = new HashMap();

		StringBuffer bodyCreateDeclaredTestMethods = new StringBuffer();

		bodyCreateDeclaredTestMethods
				.append("private static junit.framework.TestMethod[] createDeclaredTestMethods() {\n");
		bodyCreateDeclaredTestMethods.append("return new junit.framework.TestMethod[] {\n");

		boolean hasDeclaredTestMethod = false;

		boolean hasDeclaredTestSuite = false;

		for (Iterator i = pointcuts.iterator(); i.hasNext();) {
			Pointcut pointcut = (Pointcut) i.next();

			if (pointcut.acceptClass(clazz)) {

				Iterator methods = Arrays.asList(clazz.getDeclaredMethods()).iterator();
				while (methods.hasNext()) {
					CtMethod method = (CtMethod) methods.next();
					if (!instrumented.containsKey(method) && pointcut.acceptMethod(method) && pointcuts.match(method)) {

						log.debug("found method:" + clazz.getName() + "." + method.getName() + "("
								+ method.getSignature() + ")");

						modified = true;
						instrumented.put(method, null);

						if (method.getName().equals("suite")) {
							hasDeclaredTestSuite = true;
							continue;
						}

						if (hasDeclaredTestMethod) {
							bodyCreateDeclaredTestMethods.append(",");
						} else {
							hasDeclaredTestMethod = true;
						}

						countMethods++;
						/*
						 * new TestMethod("testMethodTwo") { public void
						 * run(junit.framework.TestCase testCase) throws
						 * Throwable { ((TestCase) testCase).testMethodTwo(); } }
						 */
						String nestedClassName = "_TestMethod_" + method.getName();
						CtClass nestedClass = clazz.makeNestedClass(nestedClassName, true);
						bodyCreateDeclaredTestMethods.append("new ").append(clazz.getName()).append(".").append(
								nestedClassName).append("()");
						bodyCreateDeclaredTestMethods.append(".named(\"");
						bodyCreateDeclaredTestMethods.append(method.getName());
						bodyCreateDeclaredTestMethods.append("\")\n");

						StringBuffer bodyRun = new StringBuffer();
						bodyRun.append("public void run(junit.framework.TestCase testCase) throws Throwable {\n");
						bodyRun.append("((");
						bodyRun.append(clazz.getName());
						bodyRun.append(")testCase).");
						bodyRun.append(method.getName());
						bodyRun.append("();\n");
						bodyRun.append("}\n");

						try {
							nestedClass.setSuperclass(clazz.getClassPool().get("junit.framework.TestMethod"));
							CtClass[] parameters = null;
							CtConstructor c = CtNewConstructor.make(parameters, null, nestedClass);
							// CtConstructor c = new CtConstructor(parameters,
							// nestedClass);
							nestedClass.addConstructor(c);

							nestedClass.addMethod(CtNewMethod.make(bodyRun.toString(), nestedClass));
							createdClasses.add(nestedClass);
						} catch (NotFoundException e) {
							throw new InterceptorException("Can't add nestedClass", e);
						} catch (CannotCompileException e) {
							log.info(bodyRun.toString());
							throw new InterceptorException("Can't add nestedClass", e);
						}

					}
				}
			}
		}

		bodyCreateDeclaredTestMethods.append("};}\n");

		if (modified) {
			StringBuffer code = null;
			try {
				clazz.addField(CtField.make("private static final junit.framework.TestMethod[] " + instrumentedFiled
						+ ";", clazz));
				code = bodyCreateDeclaredTestMethods;
				log.debug("bodyCreateDeclaredTestMethods:" + bodyCreateDeclaredTestMethods.toString());
				clazz.addMethod(CtNewMethod.make(bodyCreateDeclaredTestMethods.toString(), clazz));

				StringBuffer bodyDeclaredTestMethods = new StringBuffer();
				bodyDeclaredTestMethods.append("public junit.framework.TestMethod[] getDeclaredTestMethods() {\n");
				bodyDeclaredTestMethods.append("if(").append(instrumentedFiled).append(" == null) { ").append(
						instrumentedFiled).append(" = createDeclaredTestMethods();}\n");
				bodyDeclaredTestMethods.append("return ").append(instrumentedFiled).append("; }");
				code = bodyDeclaredTestMethods;

				log.debug("bodyDeclaredTestMethods:" + bodyDeclaredTestMethods.toString());

				clazz.addMethod(CtNewMethod.make(bodyDeclaredTestMethods.toString(), clazz));

				StringBuffer bodySuite = new StringBuffer();
				if (hasDeclaredTestSuite) {
					bodySuite.append("public junit.framework.Test getSuite() { return suite(); }");
				} else {
					bodySuite.append("public junit.framework.Test getSuite() { return null; }");
				}

				code = bodySuite;
				clazz.addMethod(CtNewMethod.make(bodySuite.toString(), clazz));
				code = null;
				clazz.addInterface(clazz.getClassPool().get("cldcunit.runner.TestCaseReflectionData"));

			} catch (NotFoundException e) {
				throw new InterceptorException("Can't add ReflectionData", e);
			} catch (CannotCompileException e) {
				if (code != null) {
					log.info(code.toString());
				}
				throw new InterceptorException("Can't add ReflectionData", e);
			}
			log.debug("instrumented:" + clazz.getName());
		}

		log.debug("end instrumenting:" + clazz.getName());
		if (modified) {
			return new InstrumentorResultsImpl(0, countMethods, createdClasses);
		} else {
			return InstrumentorResultsImpl.NOT_MODIFIED;
		}
	}

	public void setPointcuts(PointcutListFilter pointcuts) {
		this.pointcuts = pointcuts;
	}
}
